/*****************************************************************************
OS-DEPENDENT CONSOLE I/O FUNCTIONS

EXPORTS:
int getch(void);
int printf(const char *fmt, ...);
*****************************************************************************/
#include <stdarg.h> /* va_list, va_start(), va_end() */
#include <stdio.h> /* vsprintf() */
#include <ctype.h> /* isdigit() */

typedef struct _console
{
	unsigned char attrib, esc, esc1, esc2, esc3;
	signed char csr_x, csr_y;
} console_t;

static console_t _con =
{
	0x1F,	/* bright white on blue */
	0, 0, 0, 0, 0, 0
};
static unsigned char _vc_width = 80, _vc_height = 50;

#if defined(__TURBOC__)||defined(__DJGPP__)
#include <conio.h> /* textattr(), gotoxy(), clrscr(), putch(), getch() */
/*****************************************************************************
xxx - libc getch() uses DOS
*****************************************************************************/
#if 0
int getch(void)
{
	static short prev = -1;
/**/
	unsigned char msb, lsb;
	int ret_val;

	if(prev != -1)
	{
		ret_val = prev;
		prev = -1;
	}
	else
	{
		_AX = 0;
		asm int 0x16;
		msb = _AH;
		lsb = _AL;

		if(lsb != 0)
			ret_val = lsb;
		else
		{
			ret_val = 0;
			prev = msb;
		}
	}
	return ret_val;
}
#endif

#elif defined(__MINGW32__)

#include <windows.h>
#include <conio.h> /* putch(), getch() */
/*****************************************************************************
*****************************************************************************/
static void clrscr(void)
{
	CONSOLE_SCREEN_BUFFER_INFO info;
	DWORD num_chars, count;
	COORD xy =
	{
		0, 0
	};
	HANDLE h;

	h = GetStdHandle(STD_OUTPUT_HANDLE);
	if(!GetConsoleScreenBufferInfo(h, &info))
		return;
	num_chars = info.dwSize.X * info.dwSize.Y;
/* fill the entire screen with blanks */
	if(!FillConsoleOutputCharacter(h, (TCHAR)' ',
		num_chars, xy, &count))
			return;
/* now set the buffer's attributes accordingly */
	if(!FillConsoleOutputAttribute(h, info.wAttributes,
		num_chars, xy, &count))
			return;
/* put the cursor at (0, 0) */
	SetConsoleCursorPosition(h, xy);
}
/*****************************************************************************
*****************************************************************************/
static void gotoxy(unsigned x, unsigned y)
{
	COORD xy;
	HANDLE h;

	h = GetStdHandle(STD_OUTPUT_HANDLE);
	xy.X = x;
	xy.Y = y;
	SetConsoleCursorPosition(h, xy);
}
/*****************************************************************************
*****************************************************************************/
static void textattr(unsigned newattr)
{
	CONSOLE_SCREEN_BUFFER_INFO info;
	HANDLE h;

	h = GetStdHandle(STD_OUTPUT_HANDLE);
	info.wAttributes = newattr;
	SetConsoleTextAttribute(h, info.wAttributes);
}
#endif
/*****************************************************************************
*****************************************************************************/
static void my_textattr(int att)
{
	static const char ansi_to_vga[] =
	{
		0, 4, 2, 6, 1, 5, 3, 7
	};
/**/
	unsigned char new_att;

	new_att = _con.attrib;
/* bold on/off */
	if(att == 0)
		new_att &= ~0x08;
	else if(att == 1)
		new_att |= 0x08;
/* set foreground color */
	else if(att >= 30 && att <= 37)
	{
		att = ansi_to_vga[att - 30];
		new_att = (new_att & ~0x07) | att;
	}
/* set background color */
	else if(att >= 40 && att <= 47)
	{
		att = ansi_to_vga[att - 40] << 4;
		new_att = (new_att & ~0x70) | att;
	}
	_con.attrib = new_att;
	textattr(new_att);
}
/*****************************************************************************
*****************************************************************************/
static int my_putch(int c)
{
/* state machine to handle the escape sequences
ESC */
	if(_con.esc == 1)
	{
		if(c == '[')
		{
			_con.esc++;
			_con.esc1 = 0;
			return c;
		}
		/* else fall-through: zero esc and print c */
	}
/* ESC[ */
	else if(_con.esc == 2)
	{
		if(isdigit(c))
		{
			_con.esc1 = _con.esc1 * 10 + c - '0';
			return c;
		}
		else if(c == ';')
		{
			_con.esc++;
			_con.esc2 = 0;
			return c;
		}
/* ESC[2J -- clear screen */
		else if(c == 'J')
		{
			if(_con.esc1 == 2)
			{
				clrscr();
				_con.csr_x = _con.csr_y = 0;
			}
		}
/* ESC[num1m -- set attribute num1 */
		else if(c == 'm')
			my_textattr(_con.esc1);
		_con.esc = 0;	/* anything else with one numeric arg */
		return c;
	}
/* ESC[num1; */
	else if(_con.esc == 3)
	{
		if(isdigit(c))
		{
			_con.esc2 = _con.esc2 * 10 + c - '0';
			return c;
		}
		else if(c == ';')
		{
			_con.esc++;	/* ESC[num1;num2; */
			_con.esc3 = 0;
			return c;
		}
/* ESC[num1;num2H -- move cursor to num1,num2 */
		else if(c == 'H')
		{
			if(_con.esc2 < _vc_width)
				_con.csr_x = _con.esc2;
			if(_con.esc1 < _vc_height)
				_con.csr_y = _con.esc1;
			gotoxy(_con.csr_x, _con.csr_y);
		}
/* ESC[num1;num2m -- set attributes num1,num2 */
		else if(c == 'm')
		{
			my_textattr(_con.esc1);
			my_textattr(_con.esc2);
		}
		_con.esc = 0;
		return c;
	}
/* ESC[num1;num2;num3 */
	else if(_con.esc == 4)
	{
		if(isdigit(c))
		{
			_con.esc3 = _con.esc3 * 10 + c - '0';
			return c;
		}
/* ESC[num1;num2;num3m -- set attributes num1,num2,num3 */
		else if(c == 'm')
		{
			my_textattr(_con.esc1);
			my_textattr(_con.esc2);
			my_textattr(_con.esc3);
		}
		_con.esc = 0;
		return c;
	}
	_con.esc = 0;

/* escape character */
	if(c == 0x1B)
	{
		_con.esc = 1;
		return c;
	}
/* backspace */
	if(c == 0x08)
	{
		if(_con.csr_x != 0)
			_con.csr_x--;
	}
/* tab */
	else if(c == 0x09)
		_con.csr_x = (_con.csr_x + 8) & ~(8 - 1);
/* carriage return */
	else if(c == '\r')	/* 0x0D */
		_con.csr_x = 0;
/* line feed
	else if(c == '\n')	0x0A
		_con.csr_y++; */
/* CR/LF */
	else if(c == '\n')	/* ### - 0x0A again */
	{
		_con.csr_x = 0;
		_con.csr_y++;
	}
/* printable ASCII */
	else if(c >= ' ')
	{
		gotoxy(_con.csr_x + 1, _con.csr_y + 1);
		putch(c);
		_con.csr_x++;
	}
	if(_con.csr_x >= _vc_width)
	{
		_con.csr_x = 0;
		_con.csr_y++;
	}
	if(_con.csr_y >= _vc_height)
		_con.csr_y = _vc_height - 1;
	gotoxy(_con.csr_x + 1, _con.csr_y + 1);
	return c;
}
/*****************************************************************************
*****************************************************************************/
static int my_cputs(const char *s)
{
	int ret_val = 0;

	for(; *s != '\0'; s++)
		ret_val = my_putch(*s);
	return ret_val;
}
/*****************************************************************************
*****************************************************************************/
int printf(const char *fmt, ...)
{
	char buf[256];
	va_list args;
	int ret_val;

	va_start(args, fmt);
	ret_val = vsprintf(buf, fmt, args);
	va_end(args);
	my_cputs(buf);
	return ret_val;
}
