/*****************************************************************************
OS-DEPENDENT DISK I/O FUNCTIONS

EXPORTS:
int cache_sector(drv_t *drv, unsigned long lba, unsigned char **buf_p);
int open_drive(drv_t *drv, unsigned char dev_num, signed char part_num);
int read_disk(drv_t *drv, unsigned long sector, unsigned offset,
		unsigned count, unsigned char HUGE *buf);

For open_drive():
	dev_num = 0, 1, 2... for floppy drives
	dev_num = 0x80, 0x81, 0x82... for hard drives

	part_num = 0...3 to open one of the hard drive primary partitions
	part_num = -1u (==255) to open the entire hard drive
	part_num ignored if opening floppy drive
*****************************************************************************/
#include <string.h> /* memset() */
#include "fsdefs.h"
#include "errno.h"

/* IMPORTS
from CONIO.C */
int printf(const char *fmt, ...);

/* from FAT.C */
int check_if_fat_bootsector(unsigned char *buf);

#define	CACHE_SIZE	32768u
/* maximum and minium bytes-per-sector */
#define	MAX_BPS		4096
#define	MIN_BPS		512
/* 512 bytes is MINIMUM bytes per sector */
#define	MAX_CACHE	(CACHE_SIZE / MIN_BPS)
/*----------------------------------------------------------------------------
DOS CODE

EXPORTS:
static int alloc_cache(unsigned char **cache);
static int get_hard_disk_geometry(drv_t *drv);
static int read_sector_lba(drv_t *drv, unsigned long lba, unsigned char *buf);
static int read_sector_chs(drv_t *drv, unsigned long lba, unsigned char *buf);
----------------------------------------------------------------------------*/
#if defined(__TURBOC__)||defined(__DJGPP__)
#include <bios.h> /* _DISK_nnn, biosdisk() */

#if defined(__DJGPP__)
#include <sys/nearptr.h> /* __djgpp_conventional_base */
#include <stdlib.h> /* atexit() */
#include <dpmi.h> /* __dpmi_regs, __dpmi_int() */
#include <go32.h> /* _my_ss() */
#include <crt0.h> /* crtNNN */

#define	dos_int(N,R)	__dpmi_int(N,R)
#define	R_AX	x.ax
#define	R_BX	x.bx
#define	R_CX	x.cx
#define	R_DX	x.dx
#define	R_SI	x.si
#define	R_DS	x.ds
#define	R_FLAGS	x.flags

typedef __dpmi_regs regs_t;

#else
#include <dos.h> /* struct REGPACK, intr() */

#if __TURBOC__<0x0300
#define	_DISK_RESET	0
#define	_DISK_READ	2
#endif

#define	dos_int(N,R)	intr(N,R)
#define	R_AX	r_ax
#define	R_BX	r_bx
#define	R_CX	r_cx
#define	R_DX	r_dx
#define	R_SI	r_si
#define	R_DS	r_ds
#define	R_FLAGS	r_flags

typedef struct REGPACK regs_t;
#endif
/*****************************************************************************
memory for the disk cache must be aligned on a MAX_BPS boundary
so we don't get error 0x09 from INT 13h (DMA crossed 64K boundary).
Only floppy disk accesses use ISA DMA.
*****************************************************************************/
#if defined(__TURBOC__)
static int alloc_cache(unsigned char **cache)
{
/* pad with MAX_BPS-1 bytes to ensure alignment to MAX_BPS boundary */
	static unsigned char u_cache[CACHE_SIZE + MAX_BPS - 1];
/**/
	unsigned long cache_adr, seg;

/* form linear address of unaligned cache */
	seg = _DS;
	seg <<= 4;
	cache_adr = seg + (unsigned)u_cache;
/* align to MAX_BPS-byte boundary */
	cache_adr += MAX_BPS - 1;
	cache_adr &= -(long)MAX_BPS;
/* this is the address of the aligned cache */
	cache_adr -= seg;
	(*cache) = (unsigned char *)cache_adr;
	return 0;
}
/*****************************************************************************
grrrr....the disk cache and the command packet used by INT 13h AH=42h
must both be in conventional memory
*****************************************************************************/
#elif defined(__DJGPP__)
static unsigned char *_lba_pkt;
static int _cache_mem_selector, _lba_pkt_selector;

static void free_dos_mem(void)
{
	if(_cache_mem_selector != 0)
	{
		__dpmi_free_dos_memory(_cache_mem_selector);
		_cache_mem_selector = 0;
	}
	if(_lba_pkt_selector != 0)
	{
		__dpmi_free_dos_memory(_lba_pkt_selector);
		_lba_pkt_selector = 0;
	}
}
/*****************************************************************************
*****************************************************************************/
static int alloc_cache(unsigned char **cache)
{
	int temp;

/* enable nearptr access (fat DS)
This works under most DOS extenders, but not all */
	if(!(_crt0_startup_flags & _CRT0_FLAG_NEARPTR))
	{
		if(!__djgpp_nearptr_enable())
		{
			printf("alloc_cache: could not enable nearptr "
				"access\nTry using the 16-bit (real mode) "
				"version of this program, if available\n");
			return -1;
		}
	}
/* can't use the transfer buffer (__tb) because biosdisk() already uses it */
	temp = __dpmi_allocate_dos_memory((CACHE_SIZE + MAX_BPS - 1) / 16,
		&_cache_mem_selector);
	if(temp == -1)
ERR:	{
		printf("alloc_cache: can't allocate DOS memory\n");
		return -1;
	}
	temp <<= 4;
/* align to MAX_BPS-byte boundary */
	temp += MAX_BPS - 1;
	temp &= -(long)MAX_BPS;
	(*cache) = (unsigned char *)(temp + __djgpp_conventional_base);
/* this memory must be freed before we exit */
	atexit(free_dos_mem);
/* now allocate the command packet buffer used by INT 13h AH=42h */
	temp = __dpmi_allocate_dos_memory(1, &_lba_pkt_selector);
	if(temp == -1)
		goto ERR;
	temp <<= 4;
	_lba_pkt = (unsigned char *)(temp + __djgpp_conventional_base);
	return 0;
}
#endif
/*****************************************************************************
*****************************************************************************/
static int get_hard_disk_geometry(drv_t *drv)
{
	unsigned short cyls;
	regs_t regs;

/* use INT 13h AH=08h to count the number of hard drives */
	regs.R_AX = 0x0800;
	regs.R_DX = drv->dev_num;
	dos_int(0x13, &regs);
	if((regs.R_FLAGS & 0x0001) != 0)
	{
		printf("get_hard_disk_geometry: error getting geometry for "
			"drive 0x%02X\n", drv->dev_num);
		return -1;
	}
	if((drv->dev_num & 0x7F) >= (regs.R_DX & 0xFF))
	{
		printf("get_hard_disk_geometry: no such "
			"hard drive 0x%02X\n", drv->dev_num);
		return -1;
	}
	drv->bps = 512;
	drv->sectors = regs.R_CX & 0x3F;
	regs.R_DX >>= 8;
	drv->heads = regs.R_DX + 1;
	regs.R_CX >>= 6;
	cyls = regs.R_CX;
	if(cyls & 0x01)
		cyls |= 0x400;
	if(cyls & 0x02)
		cyls |= 0x800;
	cyls >>= 2;
	drv->part_size = (unsigned long)cyls * drv->heads * drv->sectors;
	DEBUG(printf("get_hard_disk_geometry: disk 0x%02X: CHS=%u:%u:%u, "
		"size=%lu", drv->dev_num, cyls, drv->heads,
		drv->sectors, drv->part_size);
		if(drv->use_lba)
			printf(", LBA\n");
		else
			printf("\n");
	)
	return 0;
}
/*****************************************************************************
*****************************************************************************/
static int read_sector_lba(drv_t *drv, unsigned long lba, unsigned char *buf)
{
#if defined(__TURBOC__)
/* the _lba_pkt used by DJGPP is accessed via DS
make this one static so it's also accessed via DS (instead of SS) */
	static unsigned char _lba_pkt[16];
#endif
	unsigned char tries;
	regs_t regs;

/*	lba += drv->part_start; */
	memset(_lba_pkt, 0, 16);
	_lba_pkt[0] = 16;
	write_le16(_lba_pkt + 2, 1);
#if defined(__TURBOC__)
	write_le32(_lba_pkt + 4, (unsigned long)buf);
	regs.R_DS = _DS;
	regs.R_SI = (unsigned)_lba_pkt;
#elif defined(__DJGPP__)
	{
		unsigned long adr;

		adr = (unsigned long)buf - __djgpp_conventional_base;
		write_le16(_lba_pkt + 4, adr & 0x0F);
		write_le16(_lba_pkt + 6, adr >> 4);

		adr = (unsigned long)_lba_pkt - __djgpp_conventional_base;
		regs.R_SI = adr & 0x0F;
		regs.R_DS = adr >> 4;
	}
#endif
	write_le32(_lba_pkt + 8, lba);
	regs.R_AX = 0x4200;
	regs.R_DX = drv->dev_num;
/* make 3 attempts */
	for(tries = 3; tries != 0; tries--)
	{
		dos_int(0x13, &regs);
		if((regs.R_FLAGS & 0x0001) == 0)
			return 0;
regs.R_AX >>= 8;
printf("error 0x%02X\n", regs.R_AX);
/* reset drive */
		biosdisk(_DISK_RESET,drv->dev_num,
			0, 0, 1, 1, 0);
	}
	regs.R_AX >>= 8;
	printf("read_sector_lba: error 0x%02X reading drive 0x%02X\n",
		regs.R_AX, drv->dev_num);
	return -1;
}
/*****************************************************************************
*****************************************************************************/
static int read_sector_chs(drv_t *drv, unsigned long lba, unsigned char *buf)
{
	unsigned char head, sect, tries;
	unsigned short cyl;
	int err;

/*	lba += drv->part_start; */
	if(lba == 0)
	{
		head = cyl = 0;
		sect = 1;
	}
	else
	{
		if(drv->sectors == 0 || drv->heads == 0)
		{
			printf("read_sector_chs: drive 0x%02X has "
				"unknown CHS geometry\n",
				drv->dev_num);
			return -1;
		}
		sect = lba % drv->sectors + 1;
		head = (lba / drv->sectors) % drv->heads;
		cyl = (lba / drv->sectors) / drv->heads;
	}
/* make 3 attempts */
	for(tries = 3; tries != 0; tries--)
	{
		err = biosdisk(_DISK_READ, drv->dev_num,
			head, cyl, sect, 1, buf);
		if(err == 0)
			return 0;
/* reset drive */
		biosdisk(_DISK_RESET,drv->dev_num,
			0, 0, 1, 1, 0);
	}
	printf("read_sector_chs: error 0x%02X reading drive 0x%02X\n",
		err, drv->dev_num);
	return -1;
}
/*----------------------------------------------------------------------------
WINDOWS NT CODE (UNTESTED)

EXPORTS:
static int alloc_cache(unsigned char **cache);
static int get_hard_disk_geometry(drv_t *drv);
static int read_sector_lba(drv_t *drv, unsigned long lba, unsigned char *buf);
static int read_sector_chs(drv_t *drv, unsigned long lba, unsigned char *buf);
----------------------------------------------------------------------------*/
#elif defined(__MINGW32__)
#include <windows.h>
#include <stdio.h> /* sprintf() */

/* need these for Win9x only */
#define	VWIN32_DIOC_DOS_INT13	4

typedef struct
{
	DWORD ebx, edx, ecx, eax, edi, esi, eflags;
} regs_t;

static OSVERSIONINFO _win_version;
/*****************************************************************************
xxx - must the cache be aligned? must it be in conventional memory?
if yes to either of these, how do I do it?
*****************************************************************************/
static int alloc_cache(unsigned char **cache)
{
	static unsigned char u_cache[CACHE_SIZE];
/**/

	(*cache) = u_cache;
	return 0;
}
/*****************************************************************************
*****************************************************************************/
static int get_hard_disk_geometry(drv_t *drv)
{
	if(_win_version.dwPlatformId == VER_PLATFORM_WIN32_NT)
		return 0; /* xxx - nothing to do for NT? */
	else
	{
		printf("get_hard_disk_geometry: no hard disk access "
			"via VWIN32\nIf you have Win9x, use the DOS "
			"version of this program, if available\n\n");
		return -1;
	}
}
/*****************************************************************************
*****************************************************************************/
static int read_sector_lba(drv_t *drv, unsigned long lba, unsigned char *buf)
{
	char dev_name[64];
	HANDLE disk;
	DWORD got;
	BOOL ok;

/*	lba += drv->part_start; */
	if(_win_version.dwPlatformId != VER_PLATFORM_WIN32_NT)
	{
		printf("read_sector_lba: LBA access only for hard disks, "
			"and no hard disk access\nvia VWIN32. If you have "
			"Win9x, use the DOS version of this program,\n"
			"if available\n\n");
		return -1;
	}
	if(drv->dev_num < 0x80)
		sprintf(dev_name, "\\\\.\\%c:", drv->dev_num + 'A');
	else
		sprintf(dev_name, "\\\\.\\PhysicalDrive%u",
			drv->dev_num - 0x80);
/* open the drive */
	disk = CreateFile(dev_name, 0, FILE_SHARE_READ | FILE_SHARE_WRITE,
		NULL, OPEN_EXISTING, 0, 0);
	if(disk == INVALID_HANDLE_VALUE)
ERR:	{
		printf("cache_sector_nt: error reading drive 0x%02X\n",
			drv->dev_num);
		return -1;
	}
/* xxx - before ReadFile(), must seek to byte offset (lba * drv->bps) */
	ok = ReadFile(disk, buf, drv->bps, &got, 0 /* xxx - LPOVERLAPPED */);
	CloseHandle(drv);
	if(!ok)
		goto ERR;
	return 0;
}
/*****************************************************************************
*****************************************************************************/
static int read_sector_chs(drv_t *drv, unsigned long lba,
		unsigned char *buf)
{
	unsigned char head, sect;
	unsigned short cyl;
	regs_t regs;
	HANDLE vxd;
	DWORD got;
	BOOL ok;

/* hmmmm... */
	if(_win_version.dwPlatformId == VER_PLATFORM_WIN32_NT)
		return read_sector_lba(drv, lba, buf);
/*	lba += drv->part_start; */
	if(lba == 0)
	{
		head = cyl = 0;
		sect = 1;
	}
	else
	{
		if(drv->sectors == 0 || drv->heads == 0)
		{
			printf("read_sector_chs: drive 0x%02X has "
				"unknown CHS geometry\n",
				drv->dev_num);
			return -1;
		}
		sect = lba % drv->sectors + 1;
		head = (lba / drv->sectors) % drv->heads;
		cyl = (lba / drv->sectors) / drv->heads;
	}
/* open VWIN32.VXD */
	vxd = CreateFile("\\\\.\\vwin32", 0, 0, NULL, 0,
		FILE_FLAG_DELETE_ON_CLOSE, NULL);
	if(vxd == INVALID_HANDLE_VALUE)
	{
		printf("read_sector_chs: error opening VWIN32\n");
		return -1;
	}
/* use INT 13h AH=02h to read disk */
	regs.ebx = (unsigned)buf;
	regs.ecx = cyl & 0xFF;
	regs.ecx <<= 8;
	regs.ecx |= sect;
	if(cyl & 0x0100)
		regs.ecx |= 0x40;
	if(cyl & 0x0200)
		regs.ecx |= 0x80;
	regs.edx = head;
	regs.edx <<= 8;
	regs.edx |= drv->dev_num;
	regs.eax = 0x0201;
	ok = DeviceIoControl(vxd, VWIN32_DIOC_DOS_INT13,
		&regs, sizeof(regs), &regs, sizeof(regs), &got, 0);
/* close VWIN32 */
	CloseHandle(vxd);
	if(ok && (regs.eflags & 0x0001) == 0)
		return 0;
	regs.eax >>= 8;
	printf("read_sector_chs: error 0x%02lX reading drive 0x%02X\n",
		regs.eax, drv->dev_num);
	return -1;
}
/*----------------------------------------------------------------------------
LINUX CODE

EXPORTS:
static int alloc_cache(unsigned char **cache);
static int get_hard_disk_geometry(drv_t *drv);
static int read_sector_lba(drv_t *drv, unsigned long lba, unsigned char *buf);
static int read_sector_chs(drv_t *drv, unsigned long lba, unsigned char *buf);
----------------------------------------------------------------------------*/
#elif defined(linux)
/* xxx - add Linux code */
#endif
/*----------------------------------------------------------------------------
COMMON CODE
----------------------------------------------------------------------------*/
/*****************************************************************************
*****************************************************************************/
int cache_sector(drv_t *drv, unsigned long lba, unsigned char **buf_p)
{
	static unsigned long cached_lba[MAX_CACHE];
	static unsigned char cached_dev[MAX_CACHE];
	static unsigned char *cache, evict;
/**/
	unsigned i, max;
	int err;

	if(cache == NULL)
	{
		err = alloc_cache(&cache);
		if(err != 0)
			return err;
		for(i = 0; i < MAX_CACHE; i++)
			cached_lba[i] = -1uL;
	}
	lba += drv->part_start;
	max = CACHE_SIZE / drv->bps;
/* see if this sector is cached
DUMBDRV :) */
	for(i = 0; i < max; i++)
	{
		if(cached_lba[i] == lba &&
			cached_dev[i] == drv->dev_num)
		{
			(*buf_p) = cache + i * drv->bps;
			return 0;
		}
	}
/* not cached, find a free buffer for it */
	for(i = 0; i < max; i++)
	{
		if(cached_lba[i] == -1uL)
			break;
	}
/* no free buffer, kick out someone else */
	if(i >= max)
	{
		i = evict;
		evict++;
		if(evict >= max)
			evict = 0;
	}
/* load it */
	cached_lba[i] = lba;
	cached_dev[i] = drv->dev_num;
	(*buf_p) = cache + i * drv->bps;
/* debug */
memset(*buf_p, 0x55, drv->bps);
	if(drv->use_lba)
		return read_sector_lba(drv, lba, *buf_p);
	return read_sector_chs(drv, lba, *buf_p);
}
/*****************************************************************************
put these functions here if module FAT.C is not included
*****************************************************************************/
#if !defined(FS_FAT)
static int lg2(unsigned arg)
{
	unsigned char log;

	for(log = 0; log < 16; log++)
	{
		if(arg & 1)
		{
			arg >>= 1;
			return (arg != 0) ? -1 : log;
		}
		arg >>= 1;
	}
	return -1;
}
/*****************************************************************************
*****************************************************************************/
int check_if_fat_bootsector(unsigned char *buf)
{
	unsigned short temp;
	int bad = 0;

	DEBUG(printf("check_if_fat_bootsector:\n");)
/* must start with 16-bit JMP or 8-bit JMP plus NOP */
	if(buf[0] == 0xE9)
		/* OK */;
	else if(buf[0] == 0xEB && buf[2] == 0x90)
		/* OK */;
	else
	{
		DEBUG(printf("\tMissing JMP/NOP\n");)
		bad = 1;
	}
	temp = buf[13];
	if(lg2(temp) < 0)
	{
		DEBUG(printf("\tSectors per cluster (%u) "
			"is not a power of 2\n", temp);)
		bad = 1;
	}
	temp = buf[16];
	temp = read_le16(buf + 24);
	if(temp == 0 || temp > 63)
	{
		DEBUG(printf("\tInvalid number of sectors (%u)\n", temp);)
		bad = 1;
	}
	temp = read_le16(buf + 26);
	if(temp == 0 || temp > 255)
	{
		DEBUG(printf("\tInvalid number of heads (%u)\n", temp);)
		bad = 1;
	}
	return bad;
}
#endif
/*****************************************************************************
*****************************************************************************/
static int get_fat_floppy_geometry(drv_t *drv)
{
	unsigned char *buf;
	int err;

	err = cache_sector(drv, 0, &buf);
	if(err != 0)
		return -1; /* fatal error, e.g. no disk in drive */
	if(check_if_fat_bootsector(buf) != 0)
		return 1; /* non-fatal error; try something else */
	drv->bps = read_le16(buf + 11);
	drv->sectors = read_le16(buf + 24);
	drv->heads = read_le16(buf + 26);
	drv->part_size = read_le16(buf + 19);
	if(drv->part_size == 0)
		drv->part_size = read_le32(buf + 32);
	DEBUG(printf("get_fat_floppy_geometry: disk 0x%02X: CHS=??:%u:%u, "
		"size=%lu\n", drv->dev_num, drv->heads,
		drv->sectors, drv->part_size);)
	return 0;
}
/*****************************************************************************
direct binary search to zero in on actual sectors-per-track value
this code assumes 2 heads and 80 tracks, but they could also be probed
*****************************************************************************/
static int probe_floppy_geometry(drv_t *drv)
{
	unsigned short delta;
	unsigned char *buf;
	int err;

	drv->bps = 512;
/* check if disk in drive by reading sector 0
this also clears the disk-change state */
	err = cache_sector(drv, 0, &buf);
	if(err != 0)
		return err;
/* I told you it was slow... */
	printf("probe_floppy_geometry: please wait up to 15 seconds\n");
	delta = drv->sectors = 16;
	do
	{
		delta >>= 1;
		DEBUG(printf("sectors=%d, delta=%d\n", drv->sectors, delta);)
		err = cache_sector(drv, drv->sectors - 1, &buf);
		if(err == 0)
			drv->sectors += delta;
		else
		{
			DEBUG(printf("\terror\n");)
			drv->sectors -= delta;
		}
	} while(delta != 0);
	if(err != 0)
		drv->sectors--;
/* xxx - probe number of heads, too */
	drv->heads = 2;
/* xxx - probe number of tracks, too */
	drv->part_size = 80L * drv->heads * drv->sectors;
	DEBUG(printf("probe_floppy_geometry: disk 0x%02X: CHS=??:%u:%u, "
		"size=%lu\n", drv->dev_num, drv->heads,
		drv->sectors, drv->part_size);)
	return 0;
}
/*****************************************************************************
*****************************************************************************/
static int get_floppy_disk_geometry(drv_t *drv)
{
	int err;

	err = get_fat_floppy_geometry(drv);
	if(err > 0)
	{
/* because this is so slow, try it last */
#if 0
		err = probe_floppy_geometry(drv);
#else
/* ...or just assume 1.44 meg disk */
		drv->bps = 512;
		drv->sectors = 18;
		drv->heads = 2;
		drv->part_size = 80L * drv->heads * drv->sectors;
		err = 0;
#endif
	}
	return err;
}
/*****************************************************************************
*****************************************************************************/
static int get_partition_info(drv_t *drv)
{
	unsigned char *buf, *ptab_rec;
	int err;

/* read sector 0: the partition table, or Master Boot Record (MBR) */
	err = cache_sector(drv, 0, &buf);
	if(err != 0)
		return err;
/* make sure the desired partition exists */
	ptab_rec = buf + 446 + 16 * drv->part_num;
	if(ptab_rec[4] == 0)
	{
		printf("get_partition_info: no such partition %u on drive "
			"0x%02X\n", drv->part_num, drv->dev_num);
		return -1;
	}
/* xxx - something's wrong with LBA -- debug */
if(ptab_rec[4] != 1 && ptab_rec[4] != 6 && ptab_rec[4] != 0x0B &&
ptab_rec[4] != 0x0C && ptab_rec[4] != 0x0E && ptab_rec[4] != 0x0F)
{ printf("non-FAT partition -- bad partition table or LBA problem or "
"software error or somedamnthing\n");
return -1; }

/* get part_type, part_start, part_size */
	drv->part_type = ptab_rec[4];
	drv->part_start = read_le32(ptab_rec + 8);
	drv->part_size = read_le32(ptab_rec + 12);
	DEBUG(printf("get_partition_info: disk 0x%02X, part %u: start=%lu, "
		"size=%lu, type=0x%02X\n", drv->dev_num, drv->part_num,
		drv->part_start, drv->part_size, drv->part_type);)
	return 0;
}
/*****************************************************************************
*****************************************************************************/
int open_drive(drv_t *drv, unsigned char dev_num, signed char part_num)
{
	int err = -1;

	drv->dev_num = dev_num;
	drv->part_num = part_num;
	drv->part_start = 0;
	drv->part_type = 0;
/* initial values just to make cache_sector() work */
	drv->bps = 512;
	drv->heads = 1;
	drv->sectors = 127;
/* opening floppy drive */
	if(dev_num < 0x80)
	{
		drv->use_lba = 0;
/* set drv->heads, drv->sectors, drv->part_size */
		err = get_floppy_disk_geometry(drv);
		if(err != 0)
			return err;
	}
/* opening hard drive */
	else
	{
		if(part_num < -1 || part_num > 3)
		{
			printf("open_drive: partition number (%d) must be "
				"-1 ... 3\n", part_num);
			return -1;
		}
/* set drv->heads, drv->sectors for CHS hard disk,
set drv->part_size, drv->bps for both CHS and LBA hard disks */
		err = get_hard_disk_geometry(drv);
		if(err != 0)
			return err;
/* set drv->part_size, drv->part_start, drv->part_type */
		if(part_num != -1)
			err = get_partition_info(drv);
	}
	return err;
}
/*****************************************************************************
xxx - this is slow
*****************************************************************************/
#if defined(__TURBOC__)
static void memcpy2f(unsigned char HUGE **dst, unsigned char *src,
		unsigned count)
{
	unsigned short dst_seg, dst_off;

	dst_seg = FP_SEG(*dst);
	dst_off = FP_OFF(*dst);
	for(; count != 0; count--)
	{
		pokeb(dst_seg, dst_off, *src);
		dst_off++;
		if(dst_off == 0)
			dst_seg += 0x1000;
		src++;
	}
	(*dst) = MK_FP(dst_seg, dst_off);
}
#else
static void memcpy2f(unsigned char **dst_p, unsigned char *src,
		unsigned count)
{
	unsigned char *dst;

	dst = (*dst_p);
	memcpy(dst, src, count);
	dst += count;
	(*dst_p) = dst;
}
#endif
/*****************************************************************************
read from contiguous disk sectors. Starting offset in first sector
and ending offset in last sector are completely arbitrary
*****************************************************************************/
int read_disk(drv_t *drv, unsigned long sector, unsigned offset,
		unsigned count, unsigned char HUGE *buf)
{
	unsigned char *cache;
	unsigned w;
	int err;

/* skip leading sectors */
	while(offset >= drv->bps)
	{
		sector++;
		offset -= drv->bps;
	}
/* read full or partial sectors */
	while(count != 0)
	{
		err = cache_sector(drv, sector, &cache);
		if(err != 0)
			return err;
		w = drv->bps - offset;
		if(w > count)
			w = count;
		memcpy2f(&buf, cache + offset, w);
		/* buf += w; */
		count -= w;

		sector++;
		offset = 0;
	}
	return 0;
}
